unit hiConsole;

interface

uses
  Windows, Kol, Share, Debug;

type
  THIConsole = class(TDebug)
    private
      FStdInH: THandle;
      FStdOutH: THandle;
      FStdErrH: THandle;
      FEvClose: THandle;
      
      FReadingThread: PThread;
      
      procedure WriteString(Handle: THandle; const S: string);
      
      function _ExecuteRead(Sender: PThread): Integer;
      procedure PrepareReading;
      procedure StartReading;
      procedure StopReading;
      
    public
      _prop_Icon: HICON;
      _prop_Title: string;
      _prop_WaitClose: Byte;

      _event_onStart: THI_Event;
      _event_onDataIn: THI_Event;
      _event_onDetach: THI_Event;
      
      _data_Point: THI_Event;
      _data_CtrlC: THI_Event;
      _data_Close: THI_Event;
      _data_Break: THI_Event;
      
      destructor Destroy; override;
      procedure Start;
      function ConsoleEvent(code: DWord): Bool;

      procedure _work_doWrite(var _Data: TData; Index: Word);
      procedure _work_doWriteErr(var _Data: TData; Index: Word);
      procedure _work_doPosition(var _Data: TData; Index: Word);
      procedure _work_doTextAttribute(var _Data: TData; Index: Word);
      procedure _work_doClose(var _Data: TData; Index: Word);

      procedure _var_Read(var _Data: TData; Index: Word);
  end;

implementation

const
  READ_BUFFER_SIZE = 8192;

var
  cn: THIConsole;

function CtrlHandler(fdwCtrlType: DWord): Bool; stdcall;
begin
  Result := cn.ConsoleEvent(fdwCtrlType);
end;

destructor THIConsole.Destroy;
begin
  // Free()  Terminate() ,    .
  //       Terminate().
  //          ReadFile() ,
  //    STDIN    ,     ,    .
  //  CancelSynchronousIo()    Windows Vista.
  //   Terminate()     .
  //      -        .
  FReadingThread.Free;
  
  inherited;
end;


function THIConsole._ExecuteRead(Sender: PThread): Integer;
var
  BR: DWord;
  Buf: string;
begin
  Result := 0;
  SetLength(Buf, READ_BUFFER_SIZE);
  
  // ReadFile()         "  ".
  // CloseHandle(FStdInH)    .
  // SetEvent(FStdInH)   ,   FStdInH -   input buffer 
  // (  'CONIN$').    pipe   .
  //  "Overlapped I/O"    ,       FStdInH.
  while ReadFile(FStdInH, Pointer(Buf)^, READ_BUFFER_SIZE, BR, nil) do
  begin
    if Sender.Tag <> 0 then Exit;
    //  BR = 0 , ,   Ctrl+Z   ( SetEvent(FStdInH)  'CONIN$')
    _hi_OnEvent(_event_onDataIn, Copy(Buf, 1, BR));
    
    if Sender.Tag <> 0 then Exit; //   onDataIn  doClose
  end;
  
  if Sender.Tag = 0 then _hi_OnEvent(_event_onDetach);
end;

procedure THIConsole.PrepareReading;
begin
  //   onDataIn/onDetach    
  if Assigned(_event_onDataIn.Event) or Assigned(_event_onDetach.Event) then
  begin
    _prop_WaitClose := 1;
    FReadingThread := {$ifdef F_P}NewThreadforFPC{$else}NewThread{$endif};
    FReadingThread.OnExecute := _ExecuteRead;
  end;
end;

procedure THIConsole.StartReading;
begin
  if FReadingThread <> nil then
    FReadingThread.Resume;
end;

procedure THIConsole.StopReading;
begin
  if FReadingThread <> nil then
  begin
    //       
    // (  -          ReadFile())
    FReadingThread.Tag := 1;
  end;
end;

procedure THIConsole.Start;
begin
  cn := Self;
  
  FStdInH := GetStdHandle(STD_INPUT_HANDLE);
  FStdOutH := GetStdHandle(STD_OUTPUT_HANDLE);
  FStdErrH := GetStdHandle(STD_ERROR_HANDLE);
  
  SetConsoleTitle(PChar(_prop_Title));
  SetConsoleCtrlHandler(@CtrlHandler, True);
  
  PrepareReading;
  
  if _prop_WaitClose <> 0 then
    FEvClose := CreateEvent(nil, False{ManualReset}, False, nil);
  
  EventOn;
  InitDo;
  
  _hi_OnEvent(_event_onStart);
  
  StartReading;
  
  //    doClose
  if FEvClose <> 0 then
  begin
    WaitForSingleObject(FEvClose, INFINITE);
    CloseHandle(FEvClose);
    StopReading; //     FEvClose <> 0
    Sleep(200); //    doClose 
  end;
end;

function THIConsole.ConsoleEvent;
begin
  Result := False;
  case Code of
    CTRL_C_EVENT: Result := ToIntegerEvent(_data_CtrlC) <> 0;
    CTRL_CLOSE_EVENT: Result := ToIntegerEvent(_data_Close) <> 0;
    CTRL_BREAK_EVENT: Result := ToIntegerEvent(_data_Break) <> 0;
  end;
end;

procedure THIConsole.WriteString(Handle: THandle; const S: string);
var
  BW: DWord;
begin
  WriteFile(Handle, Pointer(S)^, Length(S), BW, nil); 
end;

procedure THIConsole._work_doWrite;
begin
  WriteString(FStdOutH, ToString(_Data));
end;

procedure THIConsole._work_doWriteErr;
begin
  WriteString(FStdErrH, ToString(_Data));
end;

procedure THIConsole._work_doPosition;
var
  P: Cardinal;
begin
  P := ReadInteger(_Data, _data_Point, 0);
  SetConsoleCursorPosition(FStdOutH, Coord(P));
end;

procedure THIConsole._work_doTextAttribute;
begin
  SetConsoleTextAttribute(FStdOutH, ToInteger(_Data));
end;

procedure THIConsole._work_doClose;
begin
  //     
  SetEvent(FEvClose);
end;

procedure THIConsole._var_Read;
var
  S: string;
begin
  if FReadingThread <> nil then Exit;
  
  ReadLn(S);
  dtString(_Data, S);
end;

end.
